/**
 * This file is part of ArmarX.
 *
 * ArmarX is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * ArmarX is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 *
 * @package    Navigation::ArmarXObjects::core
 * @author     Fabian Reister ( fabian dot reister at kit dot edu )
 * @author     Christian R. G. Dreher ( c dot dreher at kit dot edu )
 * @date       2021
 * @copyright  http://www.gnu.org/licenses/gpl-2.0.txt
 *             GNU General Public License
 */

#include <vector>

#include <ArmarXCore/core/logging/Logging.h>

#include <armarx/navigation/core/Trajectory.h>
#include <armarx/navigation/core/types.h>

// test includes and other stuff
#define BOOST_TEST_MODULE Navigation::ArmarXLibraries::core
#define ARMARX_BOOST_TEST

#include <armarx/navigation/Test.h>

BOOST_AUTO_TEST_CASE(testPathLength)
{
    armarx::navigation::core::Path path{
        armarx::navigation::core::Pose(Eigen::Translation3f(0, 0, 0)),
        armarx::navigation::core::Pose(Eigen::Translation3f(0, 2000, 0)),
        armarx::navigation::core::Pose(Eigen::Translation3f(0, 4000, 0))};

    const auto traj = armarx::navigation::core::Trajectory::FromPath(path, 100);
    BOOST_CHECK_CLOSE(traj.length(), 4000, 0.01);
}

BOOST_AUTO_TEST_CASE(testResampleAlongLine)
{
    armarx::navigation::core::Path path{
        armarx::navigation::core::Pose(Eigen::Translation3f(0, 0, 0)),
        armarx::navigation::core::Pose(Eigen::Translation3f(0, 2000, 0))};

    const auto traj = armarx::navigation::core::Trajectory::FromPath(path, 100);
    BOOST_CHECK_EQUAL(traj.points().size(), 2);

    const auto resampledTraj = traj.resample(500);

    for (const auto& pt : resampledTraj.positions())
    {
        ARMARX_DEBUG << VAROUT(pt);
    }

    BOOST_CHECK_EQUAL(resampledTraj.points().size(), 4);
}

BOOST_AUTO_TEST_CASE(testResampleAlongLineWithWaypoint)
{
    armarx::navigation::core::Path path{
        armarx::navigation::core::Pose(Eigen::Translation3f(0, 0, 0)),
        armarx::navigation::core::Pose(Eigen::Translation3f(0, 1050, 0)),
        armarx::navigation::core::Pose(Eigen::Translation3f(0, 2100, 0))};

    const auto traj = armarx::navigation::core::Trajectory::FromPath(path, 100);
    BOOST_CHECK_EQUAL(traj.points().size(), 3);

    const auto resampledTraj = traj.resample(500);

    for (const auto& pt : resampledTraj.positions())
    {
        ARMARX_DEBUG << VAROUT(pt);
    }

    BOOST_CHECK_EQUAL(resampledTraj.points().size(), 5);
}
