#include "OrientationOptimizer.h"

#include <range/v3/range/conversion.hpp>
#include <range/v3/view/drop.hpp>
#include <range/v3/view/drop_last.hpp>
#include <range/v3/view/transform.hpp>
#include <range/v3/view/zip.hpp>

#include <ceres/ceres.h>
#include <ceres/cost_function.h>

#include <SimoxUtility/math/convert/mat4f_to_xyyaw.h>
#include <SimoxUtility/math/convert/rpy_to_mat3f.h>

#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <armarx/navigation/core/Trajectory.h>


namespace armarx::navigation::glob_plan
{
    OrientationOptimizer::OrientationOptimizer(const core::Trajectory& trajectory,
                                               const Params& params) :
        trajectory(trajectory), params(params)
    {
    }

    //   problem.AddResidualBlock(new MyBinaryCostFunction(...), nullptr, x2, x1);


    struct CostFunctor
    {
        template <typename T>
        bool
        operator()(const T* const x, T* residual) const
        {
            residual[0] = 10.0 - x[0];
            return true;
        }
    };

    struct SmoothOrientationCostFunctor
    {

        SmoothOrientationCostFunctor(const double weight) : weight(weight)
        {
        }

        template <typename T>
        bool
        operator()(const T* const yawPrev,
                   const T* const yaw,
                   const T* const yawNext,
                   T* residual) const
        {
            // TODO periodic
            residual[0] = weight * (yawPrev[0] - yaw[0]);
            residual[1] = weight * (yawNext[0] - yaw[0]);
            return true;
        }

    private:
        const double weight;
    };

    struct SmoothOrientationFixedPreCostFunctor
    {

        SmoothOrientationFixedPreCostFunctor(const double yawPre, const double weight) :
            yawPre(yawPre), weight(weight)
        {
        }

        template <typename T>
        bool
        operator()(const T* const yaw, const T* const yawNext, T* residual) const
        {
            // TODO periodic
            residual[0] = weight * (yawPre - yaw[0]);
            residual[1] = weight * (yawNext[0] - yaw[0]);
            return true;
        }

    private:
        const double yawPre;
        const double weight;
    };

    struct SmoothOrientationFixedNextCostFunctor
    {

        SmoothOrientationFixedNextCostFunctor(const double yawNext, const double weight) :
            yawNext(yawNext), weight(weight)
        {
        }

        template <typename T>
        bool
        operator()(const T* const yawPre, const T* const yaw, T* residual) const
        {
            // TODO periodic
            residual[0] = weight * (yawPre[0] - yaw[0]);
            residual[1] = weight * (yawNext - yaw[0]);
            return true;
        }

    private:
        const double yawNext;
        const double weight;
    };

    struct OrientationPriorCostFunctor
    {

        OrientationPriorCostFunctor(const double prior, const double weight) :
            prior(prior), weight(weight)
        {
        }

        template <typename T>
        bool
        operator()(const T* const x, T* residual) const
        {
            // TODO periodic
            residual[0] = weight * (prior - x[0]);
            return true;
        }

    private:
        const double prior;
        const double weight;
    };

    OrientationOptimizer::OptimizationResult
    OrientationOptimizer::optimize()
    {

        namespace r = ::ranges;
        namespace rv = ::ranges::views;

        const auto toYaw = [](const core::TrajectoryPoint& pt) -> double
        { return simox::math::mat4f_to_xyyaw(pt.waypoint.pose.matrix()).z(); };

        std::vector<double> orientations =
            trajectory.points() | ranges::views::transform(toYaw) | ranges::to_vector;

        ARMARX_INFO << "orientations before: " << orientations;

        // const float yawStart = orientations.front();
        // const float yawEnd = orientations.back();

        // //
        // const auto yawsVariable =
        //     orientations | ranges::views::drop(1) | ranges::views::drop_last(1) | ranges::to_vector;

        ceres::Problem problem;

        ARMARX_INFO << orientations.size() - 2 << "orientations to optimize";

        // TODO https://ceres-solver.googlesource.com/ceres-solver/+/master/examples/slam/pose_graph_2d/pose_graph_2d.cc
        //     ceres::LocalParameterization* angle_local_parameterization =
        //   ceres::AngleLocalParameterization::Create();

        // prior for all waypoints
        for (size_t i = 1; i < (orientations.size() - 1); i++)
        {
            ceres::CostFunction* priorCostFunction =
                new ceres::AutoDiffCostFunction<OrientationPriorCostFunctor, 1, 1>(
                    new OrientationPriorCostFunctor(orientations.at(i), params.priorWeight));

            problem.AddResidualBlock(priorCostFunction, nullptr, &orientations.at(i));
        }

        // smooth waypoint orientation, no start and goal nodes involved!
        for (size_t i = 2; i < (orientations.size() - 2); i++)
        {
            ceres::CostFunction* smoothCostFunction =
                new ceres::AutoDiffCostFunction<SmoothOrientationCostFunctor, 2, 1, 1, 1>(
                    new SmoothOrientationCostFunctor(params.smoothnessWeight));

            problem.AddResidualBlock(smoothCostFunction,
                                     nullptr,
                                     &orientations.at(i - 1),
                                     &orientations.at(i),
                                     &orientations.at(i + 1));
        }

        // smooth waypoint orientation, involving start
        {
            ceres::CostFunction* smoothCostFunction =
                new ceres::AutoDiffCostFunction<SmoothOrientationFixedPreCostFunctor, 2, 1, 1>(
                    new SmoothOrientationFixedPreCostFunctor(orientations.front(),
                                                             params.smoothnessWeightStartGoal));

            problem.AddResidualBlock(
                smoothCostFunction, nullptr, &orientations.at(1), &orientations.at(2));
        }

        // smooth waypoint orientation, involving goal
        {
            ceres::CostFunction* smoothCostFunction =
                new ceres::AutoDiffCostFunction<SmoothOrientationFixedNextCostFunctor, 2, 1, 1>(
                    new SmoothOrientationFixedNextCostFunctor(orientations.back(),
                                                              params.smoothnessWeightStartGoal));

            problem.AddResidualBlock(smoothCostFunction,
                                     nullptr,
                                     &orientations.at(orientations.size() - 3),
                                     &orientations.at(orientations.size() - 2));
        }


        // Run the solver!
        ceres::Solver::Options options;
        options.linear_solver_type = ceres::DENSE_QR;
        options.minimizer_progress_to_stdout = true;
        ceres::Solver::Summary summary;
        Solve(options, &problem, &summary);

        std::cout << summary.BriefReport() << "\n";

        ARMARX_INFO << "orientations after: " << orientations;

        const auto applyOrientation = [](const auto& p) -> core::TrajectoryPoint
        {
            core::TrajectoryPoint tp = p.first;
            const float yaw = p.second;

            tp.waypoint.pose.linear() = simox::math::rpy_to_mat3f(0.F, 0.F, yaw);

            return tp;
        };

        // TODO(fabian.reister): could also be in-place
        const auto modifiedTrajectory = rv::zip(trajectory.points(), orientations) |
                                        rv::transform(applyOrientation) | r::to_vector;

        return OrientationOptimizer::OptimizationResult{.trajectory = modifiedTrajectory};
    }


} // namespace armarx::navigation::glob_plan
